package buildcraft.silicon;

import java.util.EnumMap;
import java.util.HashMap;
import java.util.Map;

import io.netty.buffer.ByteBuf;
import gnu.trove.map.hash.TObjectIntHashMap;

import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.init.Items;
import net.minecraft.inventory.IInventory;
import net.minecraft.inventory.ISidedInventory;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.EnumFacing;

import buildcraft.BuildCraftSilicon;
import buildcraft.api.core.BCLog;
import buildcraft.api.core.IInvSlot;
import buildcraft.BuildCraftCore;
import buildcraft.core.lib.block.TileBuildCraft;
import buildcraft.core.lib.inventory.InventoryIterator;
import buildcraft.core.lib.inventory.SimpleInventory;
import buildcraft.core.lib.inventory.StackHelper;
import buildcraft.core.lib.utils.NBTUtils;

public class TilePackager extends TileBuildCraft implements ISidedInventory {
    private class Requirement {
        public final IInventory location;
        public final int slot;

        public Requirement(IInventory location, int slot) {
            this.location = location;
            this.slot = slot;
        }

        public boolean isValid() {
            return location.func_70302_i_() > slot && (!(location instanceof TileEntity) || !((TileEntity) location).func_145837_r());
        }

        public ItemStack getStack() {
            return location.func_70301_a(slot);
        }

        public ItemStack decrStackSize(int amount) {
            return location.func_70298_a(slot, amount);
        }

        @Override
        public boolean equals(Object other) {
            if (other == null || !(other instanceof Requirement)) {
                return false;
            }

            Requirement r = (Requirement) other;

            return r.location.equals(this.location) && r.slot == this.slot;
        }

        @Override
        public int hashCode() {
            return location.hashCode() + (slot * 17);
        }
    }

    // Slot 10 is currently missing. Left in for backwards compat.
    private static final int[] SLOTS = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 11 };

    public SimpleInventory inventoryPublic = new SimpleInventory(12, "Packager", 64);
    public SimpleInventory inventoryPattern = new SimpleInventory(9, "Packager", 64);

    private Requirement[] requirements = new Requirement[9];
    private int patternsSet;
    private int updateTime = BuildCraftCore.random.nextInt(5);

    public boolean isPatternSlotSet(int p) {
        return (patternsSet & (1 << p)) != 0;
    }

    public void setPatternSlot(int p, boolean v) {
        if (v) {
            patternsSet |= 1 << p;
        } else {
            patternsSet &= ~(1 << p);
        }
    }

    @Override
    public void func_73660_a() {
        super.func_73660_a();

        if (field_145850_b.field_72995_K) {
            return;
        }

        if ((updateTime++) % 5 == 0 && patternsSet > 0) {
            attemptCrafting(inventoryPublic.func_70301_a(9));
        }
    }

    private boolean validMissing(Requirement r, int missingCount) {
        ItemStack inputStack = r.getStack();
        if (inputStack != null && inputStack.field_77994_a >= missingCount) {
            // Check if same type used elsewhere
            for (int j = 0; j < 9; j++) {
                ItemStack comparedStack = inventoryPattern.func_70301_a(j);
                if (isPatternSlotSet(j) && comparedStack != null && StackHelper.isMatchingItem(inputStack, comparedStack, true, false)) {
                    return false;
                }
            }
            return true;
        } else {
            return false;
        }
    }

    private boolean attemptCrafting(ItemStack input) {
        // STEP 0: Make sure the conditions are correct.
        if (inventoryPublic.func_70301_a(11) != null) {
            return false;
        }

        if (input == null || input.field_77994_a == 0 || !(input.func_77973_b() == Items.field_151121_aF || input.func_77973_b() instanceof ItemPackage)) {
            return false;
        }

        if (input.func_77973_b() instanceof ItemPackage) {
            NBTTagCompound inputTag = NBTUtils.getItemData(input);

            for (int i = 0; i < 9; i++) {
                if (isPatternSlotSet(i) && inputTag.func_74764_b("item" + i)) {
                    return false;
                }
            }
        }

        TObjectIntHashMap<Requirement> reqCounts = new TObjectIntHashMap<Requirement>(9);
        int missingCount = 0;

        int filteredReqsToFulfill = 0;

        // STEP 1: Verify all requirements and nullify ones which don't match.
        // Also, add them to a Multimap so we can know which ones get used how often.
        for (int i = 0; i < 9; i++) {
            if (isPatternSlotSet(i)) {
                ItemStack inputStack = inventoryPattern.func_70301_a(i);
                if (inputStack != null) {
                    filteredReqsToFulfill++;
                } else {
                    missingCount++;
                    requirements[i] = null;
                    continue;
                }

                Requirement r = requirements[i];
                if (r == null) {
                    continue;
                }
                if (!r.isValid()) {
                    requirements[i] = null;
                    continue;
                }
                if (r.getStack() == null) {
                    requirements[i] = null;
                    continue;
                }
                if (inputStack != null) {
                    if (!StackHelper.isMatchingItem(inputStack, r.getStack(), true, false)) {
                        requirements[i] = null;
                        continue;
                    }
                }
                reqCounts.adjustOrPutValue(requirements[i], 1, 1);
                filteredReqsToFulfill--;
            } else {
                requirements[i] = null;
            }
        }

        // STEP 2: Verify that the counts are correct.
        for (Requirement r : reqCounts.keys(new Requirement[reqCounts.size()])) {
            if (r.getStack().field_77994_a < reqCounts.get(r)) {
                int allowedAmount = 0;
                for (int i = 0; i < 9; i++) {
                    if (requirements[i] != null && requirements[i].equals(r)) {
                        allowedAmount--;
                        if (allowedAmount < 0) {
                            requirements[i] = null;
                            filteredReqsToFulfill++;
                        }
                    }
                }
                reqCounts.remove(r);
            }
        }

        // STEP 3: Look for all filtered slots. We also use adjacent inventories for this.
        // STEP 3a: Local
        if (filteredReqsToFulfill > 0) {
            for (int i = 0; i < 9; i++) {
                if (filteredReqsToFulfill == 0) {
                    break;
                }
                if (isPatternSlotSet(i) && requirements[i] == null) {
                    ItemStack inputStack = inventoryPattern.func_70301_a(i);
                    if (inputStack != null) {
                        for (int j = 0; j < 9; j++) {
                            ItemStack comparedStack = inventoryPublic.func_70301_a(j);
                            if (comparedStack == null) {
                                continue;
                            }
                            Requirement r = new Requirement(this, j);
                            if (comparedStack.field_77994_a <= reqCounts.get(r)) {
                                continue;
                            }

                            if (StackHelper.isMatchingItem(inputStack, comparedStack, true, false)) {
                                requirements[i] = r;
                                filteredReqsToFulfill--;
                                reqCounts.adjustOrPutValue(r, 1, 1);
                                break;
                            }
                        }
                    }
                }
            }
        }

        // STEP 3b: Remote
        Map<EnumFacing, IInventory> invs = new HashMap<EnumFacing, IInventory>();
        if (filteredReqsToFulfill > 0 || missingCount > 0) {
            for (int i = 2; i < 6; i++) {
                TileEntity neighbor = getTile(EnumFacing.func_82600_a(i));
                if (neighbor instanceof IInventory) {
                    invs.put(EnumFacing.func_82600_a(i), (IInventory) neighbor);
                }
            }
        }

        if (filteredReqsToFulfill > 0) {
            for (EnumFacing dir : invs.keySet()) {
                if (filteredReqsToFulfill == 0) {
                    break;
                }
                IInventory inv = invs.get(dir);
                Iterable<IInvSlot> iterator = InventoryIterator.getIterable(inv, dir.func_176734_d());
                for (IInvSlot slot : iterator) {
                    if (filteredReqsToFulfill == 0) {
                        break;
                    }
                    ItemStack comparedStack = slot.getStackInSlot();
                    if (comparedStack == null || !slot.canTakeStackFromSlot(comparedStack)) {
                        continue;
                    }
                    Requirement r = new Requirement(inv, slot.getIndex());
                    if (comparedStack.field_77994_a <= reqCounts.get(r)) {
                        continue;
                    }

                    for (int j = 0; j < 9; j++) {
                        ItemStack inputStack = inventoryPattern.func_70301_a(j);
                        if (isPatternSlotSet(j) && requirements[j] == null && inputStack != null) {
                            if (StackHelper.isMatchingItem(inputStack, comparedStack, true, false)) {
                                filteredReqsToFulfill--;
                                requirements[j] = r;
                                reqCounts.adjustOrPutValue(r, 1, 1);
                                break;
                            }
                        }
                    }
                }
            }
        }

        if (filteredReqsToFulfill > 0) {
            return false;
        }

        // STEP 4: Find a matching missing.
        boolean foundMissing = false;

        if (missingCount > 0) {
            for (int i = 0; i < 9; i++) {
                Requirement r = new Requirement(this, i);
                if (reqCounts.contains(r)) {
                    continue;
                }
                if (validMissing(r, missingCount)) {
                    foundMissing = true;
                    for (int j = 0; j < 9; j++) {
                        if (requirements[j] == null && isPatternSlotSet(j) && inventoryPattern.func_70301_a(j) == null) {
                            requirements[j] = r;
                        }
                    }
                    reqCounts.adjustOrPutValue(r, missingCount, missingCount);
                    missingCount = 0;
                    break;
                }
            }
        }

        if (missingCount > 0) {
            for (EnumFacing dir : invs.keySet()) {
                if (foundMissing) {
                    break;
                }
                IInventory inv = invs.get(dir);
                Iterable<IInvSlot> iterator = InventoryIterator.getIterable(inv, dir.func_176734_d());
                for (IInvSlot slot : iterator) {
                    if (foundMissing) {
                        break;
                    }
                    Requirement r = new Requirement(inv, slot.getIndex());
                    if (reqCounts.contains(r)) {
                        continue;
                    }
                    if (validMissing(r, missingCount)) {
                        foundMissing = true;
                        for (int j = 0; j < 9; j++) {
                            if (requirements[j] == null && isPatternSlotSet(j) && inventoryPattern.func_70301_a(j) == null) {
                                requirements[j] = r;
                            }
                        }
                        reqCounts.adjustOrPutValue(r, missingCount, missingCount);
                        missingCount = 0;
                        break;
                    }
                }
            }
        }

        if (missingCount > 0) {
            return false;
        }

        // STEP 5: Craft and output.
        ItemStack pkg;
        if (input.func_77973_b() instanceof ItemPackage) {
            pkg = input.func_77946_l();
        } else {
            pkg = new ItemStack(BuildCraftSilicon.packageItem);
        }
        NBTTagCompound pkgTag = NBTUtils.getItemData(pkg);

        boolean broken = false;

        for (int i = 0; i < 9; i++) {
            if (isPatternSlotSet(i)) {
                if (requirements[i] == null) {
                    BCLog.logger.error("(Recipe Packager) At " + func_174877_v().toString() + " requirement " + i + " was null! THIS SHOULD NOT HAPPEN!");
                    broken = true;
                    continue;
                }
                ItemStack usedStack = requirements[i].decrStackSize(1);
                if (usedStack == null) {
                    BCLog.logger.error("(Recipe Packager) At " + func_174877_v().toString() + " stack at slot " + i
                        + " was too small! THIS SHOULD NOT HAPPEN!");
                    broken = true;
                    continue;
                }
                NBTTagCompound itemTag = new NBTTagCompound();
                usedStack.func_77955_b(itemTag);
                pkgTag.func_74782_a("item" + i, itemTag);
            }
        }

        if (broken) {
            return false;
        }

        ItemPackage.update(pkg);

        func_70298_a(9, 1);
        func_70299_a(11, pkg);

        return true;
    }

    @Override
    public void readData(ByteBuf stream) {
        super.readData(stream);
        patternsSet = stream.readUnsignedShort();
    }

    @Override
    public void writeData(ByteBuf stream) {
        super.writeData(stream);
        stream.writeShort(patternsSet);
    }

    @Override
    public void func_145839_a(NBTTagCompound cpd) {
        super.func_145839_a(cpd);
        inventoryPublic.readFromNBT(cpd, "items");
        inventoryPattern.readFromNBT(cpd, "pattern");
        patternsSet = cpd.func_74765_d("patternSet");
    }

    @Override
    public void func_145841_b(NBTTagCompound cpd) {
        super.func_145841_b(cpd);
        inventoryPublic.writeToNBT(cpd, "items");
        inventoryPattern.writeToNBT(cpd, "pattern");
        cpd.func_74777_a("patternSet", (short) patternsSet);
    }

    @Override
    public int func_70302_i_() {
        return inventoryPublic.func_70302_i_();
    }

    @Override
    public ItemStack func_70301_a(int slot) {
        return inventoryPublic.func_70301_a(slot);
    }

    @Override
    public ItemStack func_70298_a(int slot, int amount) {
        return inventoryPublic.func_70298_a(slot, amount);
    }

    @Override
    public ItemStack func_70304_b(int slot) {
        return inventoryPublic.func_70304_b(slot);
    }

    @Override
    public void func_70299_a(int slot, ItemStack stack) {
        inventoryPublic.func_70299_a(slot, stack);
    }

    @Override
    public String getInventoryName() {
        return "Packager";
    }

    @Override
    public int func_70297_j_() {
        return 64;
    }

    @Override
    public boolean func_70300_a(EntityPlayer player) {
        return inventoryPublic.func_70300_a(player);
    }

    @Override
    public void func_174889_b(EntityPlayer player) {
        inventoryPublic.func_174889_b(player);
    }

    @Override
    public void func_174886_c(EntityPlayer player) {
        inventoryPublic.func_174886_c(player);
    }

    @Override
    public boolean func_94041_b(int slot, ItemStack stack) {
        if (slot == 9) {
            return stack == null || stack.func_77973_b() == Items.field_151121_aF || stack.func_77973_b() instanceof ItemPackage;
        }
        return inventoryPublic.func_94041_b(slot, stack);
    }

    @Override
    public int[] func_180463_a(EnumFacing face) {
        return SLOTS;
    }

    @Override
    public boolean func_180462_a(int slot, ItemStack stack, EnumFacing side) {
        if (side.func_176740_k() != EnumFacing.Axis.Y) {
            if (slot >= 9) {
                return false;
            }
            ItemStack slotStack = inventoryPublic.func_70301_a(slot);
            if (StackHelper.canStacksMerge(stack, slotStack)) {
                return true;
            }

            for (int i = 0; i < 9; i++) {
                if (isPatternSlotSet(i)) {
                    ItemStack inputStack = inventoryPattern.func_70301_a(i);
                    if (inputStack == null) {
                        return true;
                    }
                    if (StackHelper.isMatchingItem(inputStack, stack, true, false)) {
                        return true;
                    }
                }
            }
            return false;
        } else {
            return slot == 9;
        }
    }

    @Override
    public boolean func_180461_b(int slot, ItemStack stack, EnumFacing side) {
        return slot == 11;
    }
}
